import os
import shutil
import unittest

import pkg_resources

from scrambler.command import main

SCRAMBLER_TMP = '/tmp/scrambler_tests'


class ScramblerTestCase(unittest.TestCase):

    def setUp(self):
        self.start_dir = os.getcwd()
        super(ScramblerTestCase, self).setUp()
        os.makedirs(SCRAMBLER_TMP)
        os.chdir(SCRAMBLER_TMP)
        
    def tearDown(self):
        super(ScramblerTestCase, self).tearDown()
        os.chdir(self.start_dir)
        shutil.rmtree(SCRAMBLER_TMP)
        
    def test_scrambler_command(self):
        main()
        self.assertEqual(
            os.listdir(SCRAMBLER_TMP), ['scrambled'])

        ws = pkg_resources.working_set

        expected = []
        for name, dist in ws.by_key.items(): 
            top_level = sorted(list(dist._get_metadata('top_level.txt')))
            expected += top_level
        expected = set(expected)

        linked = []
        for link in os.listdir(os.path.join(SCRAMBLER_TMP, 'scrambled')):
            if link.endswith(".py"):
                linked.append(link[:-3])
            else:
                linked.append(link)

        self.assertEqual(
            sorted(linked),
            sorted(expected))

    def test_scrambler_command_custom_dir(self):
        main(['--target', 'custom_scrambled'])
        self.assertEqual(
            os.listdir(SCRAMBLER_TMP), ['custom_scrambled'])

        ws = pkg_resources.working_set

        expected = []
        for name, dist in ws.by_key.items(): 
            top_level = sorted(list(dist._get_metadata('top_level.txt')))
            expected += top_level
        expected = set(expected)

        linked = []
        for link in os.listdir(os.path.join(SCRAMBLER_TMP, 'custom_scrambled')):
            if link.endswith(".py"):
                linked.append(link[:-3])
            else:
                linked.append(link)
        
        self.assertEqual(
            sorted(linked),
            sorted(expected))
        

    def test_scrambler_command_namespaced(self):
        main()

        ws = pkg_resources.working_set        

        namespaces = {}
        for name, dist in ws.by_key.items():
            for line in dist._get_metadata('namespace_packages.txt'):
                ns = line.split('.')[0]
                if not ns in namespaces:
                    namespaces[ns] = []
                namespaces[ns].append(dist)

        for namespace, dists in namespaces.items():
            linked = sorted(
                [x for x in os.listdir(
                    os.path.join(SCRAMBLER_TMP, 'scrambled', namespace))
                 if os.path.isdir(os.path.join(SCRAMBLER_TMP, 'scrambled', namespace, x))
                 and not x.startswith('__')])

            expected = []
            for dist in dists:
                expected += [
                    x for x in os.listdir(os.path.join(dist.location, namespace))
                    if os.path.isdir(os.path.join(dist.location, namespace, x))
                    and not x.startswith("__")]

            expected = sorted(set(expected))
            self.assertEqual(linked, expected)



"""
scrambler
"""

import os

from setuptools import setup, find_packages

version = "0.0.4"


install_requires = ['distribute']

# collective.recipe.omelette is only installed as it provides a
# good example of a namespaced package
tests_require = install_requires + ['collective.recipe.omelette']


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

long_description = read("README.rst")

setup(
    name='scrambler',
    version=version,
    description=(
        "Symlink (namespaced) python packages into a "
        + "single folder to aid development"),
    long_description=long_description,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ],
    keywords='',
    author='Ryan Northey',
    author_email='ryan@3ca.org.uk',
    url='http://github.com/phlax/scrambler',
    license='GPL',
    packages=find_packages(),
    include_package_data=True,
    test_suite="scrambler.tests",
    tests_require=tests_require,
    zip_safe=False,
    install_requires=install_requires,
    entry_points={
        'console_scripts': [
            'scrambler = scrambler.command:main']})



# CRIBBED FROM
# https://github.com/collective/collective.recipe.omelette/blob/master/collective/recipe/omelette/__init__.py

import os
import sys
import shutil
import pkg_resources
import argparse

import logging
log = logging.getLogger("scrambler")


WIN32 = False
if sys.platform[:3].lower() == "win":
    WIN32 = True

if WIN32:
    import ntfsutils.junction

    islink = ntfsutils.junction.isjunction

    def symlink(source, link_name):
        if not os.path.isdir(source):
            return
        ntfsutils.junction.create(source, link_name)

    def unlink(path):
        if not ntfsutils.junction.isjunction(path):
            return False
        ntfsutils.junction.unlink(path)
        return True

    def rmtree(location, nonlinks=True):
        # Explicitly unlink all junction'd links
        names = os.listdir(location)
        for dir in names:
            path = os.path.join(location, dir)
            if unlink(path):
                continue
            if os.path.isdir(path):
                rmtree(path)
        # Then get rid of everything else
        if nonlinks:
            shutil.rmtree(location)

else:
    symlink = os.symlink
    islink = os.path.islink
    rmtree = shutil.rmtree
    unlink = None


def create_namespaces(dist, namespaces, location, ns_base=()):
    iterator = namespaces.items

    if hasattr(namespaces, "iteritems"):
        # python 2
        iterator = namespaces.iteritems

    for k, v in iterator():
        ns_parts = ns_base + (k,)
        link_dir = os.path.join(location, *ns_parts)

        if not os.path.exists(link_dir):
            os.makedirs(link_dir)

        if len(v) > 0:
            create_namespaces(dist, v, location, ns_parts)
        egg_ns_dir = os.path.join(dist.location, *ns_parts)

        if not os.path.isdir(egg_ns_dir):
            log.info(
                "(While processing egg %s) Package '%s' is zipped. "
                "Skipping." % (
                    dist.project_name, os.path.sep.join(ns_parts)))
            continue
        dirs = os.listdir(egg_ns_dir)
        for name in dirs:
            if name.startswith('.'):
                continue
            name_parts = ns_parts + (name,)
            src = os.path.join(dist.location, *name_parts)
            dst = os.path.join(location, *name_parts)
            if os.path.exists(dst):
                continue
            symlink(src, dst)


def main(args=None):
    location = os.path.abspath('scrambled')

    parser = argparse.ArgumentParser(
        prog="scrambler",
        description='scrambler symlinks to python packages into a folder')

    parser.add_argument(
        "--target",
        help="target folder")

    # when called through 'python setup.py test' we have to ignore the
    # first passed argument
    if sys.argv[1:] == ['test']:
        args = args or []
    else:
        args = args or sys.argv[1:] or []

    parsed = parser.parse_args(args)

    if parsed.target:
        location = os.path.abspath(parsed.target)

    if not os.path.exists(location):
        os.makedirs(location)

    for dist in pkg_resources.working_set.by_key.values():
        project_name = dist.project_name
        namespaces = {}
        for line in dist._get_metadata('namespace_packages.txt'):
            ns = namespaces
            for part in line.split('.'):
                ns = ns.setdefault(part, {})

        top_level = sorted(list(dist._get_metadata('top_level.txt')))
        create_namespaces(dist, namespaces, location)

        for package_name in top_level:
            if package_name in namespaces:
                # These are processed in create_namespaces
                continue
            else:
                if not os.path.isdir(dist.location):
                    info_message = (
                        "(While processing egg %s) Package '%s' is "
                        "zipped. Skipping.")
                    log.info(
                        info_message % (
                            project_name, package_name))
                    continue
                package_location = os.path.join(dist.location, package_name)
                link_location = os.path.join(location, package_name)
                # check for single python module
                if not os.path.exists(package_location):
                    package_location = os.path.join(
                        dist.location, package_name + ".py")
                    link_location = os.path.join(
                        location, package_name + ".py")
                # check for native libs
                # XXX - this should use native_libs from above
                if not os.path.exists(package_location):
                    package_location = os.path.join(
                        dist.location, package_name + ".so")
                    link_location = os.path.join(
                        location, package_name + ".so")
                if not os.path.exists(package_location):
                    package_location = os.path.join(
                        dist.location, package_name + ".dll")
                    link_location = os.path.join(
                        location, package_name + ".dll")
                if not os.path.exists(package_location):
                    log.warn(
                        "Warning: (While processing egg %s) Package '%s' "
                        "not found. Skipping." % (
                            project_name, package_name))
                    continue

                if not os.path.exists(link_location):
                    if WIN32 and not os.path.isdir(package_location):
                        log.warn(
                            "Warning: (While processing egg %s) Can't link "
                            "files on Windows (%s -> %s). Skipping." % (
                                project_name, package_location, link_location))
                        continue
                    try:
                        symlink(package_location, link_location)
                    except OSError as e:
                        # TODO: deal with dangling symlinks
                        warning = (
                            "While processing egg %s) symlink fails "
                            "(%s, %s). Skipping.\nOriginal Exception:\n%s")
                        log.warn(
                            warning % (
                                project_name, package_location,
                                link_location, str(e)))


if __name__ == "__main__":
    main(sys.argv[1:])
